///File : Printing.cpp
//------------------------------------------------------------------------------
/*
    This file is part of DevisMaker.

    DevisMaker is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DevisMaker is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DevisMaker.  If not, see <http://www.gnu.org/licenses/>.
*/
//------------------------------------------------------------------------------
#include "Printing.hpp"
//------------------------------------------------------------------------------
void DevisPrintout::PrepareTVATable()
{
    TableClear(m_TVATbl);

    m_TVATbl.doDrawVLine = true;

    wxDC *dc = GetDC();

    StructTblLine tmpLine;
    StructTblCase tmpCase;

    // Paramétrage des lignes :
    tmpLine.isHline = false; // Pas de ligne au dessus du titre
    tmpLine.interLineSpace = m_generalCfg->GetInt("interLineTVARecapTblSpace");

    // Paramétrage des cases :
    tmpCase.extraVMarging = 0;
    tmpCase.hAlign = mdo::hAlign_Top;
    tmpCase.vAlign = mdo::vAlign_Center;

    // Case titre :
    tmpCase.font = &m_fonts.tableTitle;
	tmpCase.value = _("T");
	tmpLine.cases.push_back(tmpCase);
	tmpCase.value = _("TVA (%)");
	tmpLine.cases.push_back(tmpCase);
	tmpCase.value = _("Base");
	tmpLine.cases.push_back(tmpCase);
	tmpCase.value = _("Total");
	tmpLine.cases.push_back(tmpCase);

	m_TVATbl.lines.push_back(tmpLine); // Ajout au tableau...
	tmpLine.cases.clear(); // On videnge pour la prochaine ligne...

	// Case data
	tmpCase.font = &m_fonts.tableItem;
	tmpLine.isHline = true;

    std::vector<StructTVALine> stdTVATbl = m_DevisManager->GetTVATbl();

    for(int i = 0; i < stdTVATbl.size(); i++)
    {
        tmpCase.vAlign = mdo::vAlign_Center;
        tmpCase.value = StrStdToWx(stdTVATbl[i].t);
        tmpLine.cases.push_back(tmpCase);

        tmpCase.vAlign = mdo::vAlign_Right; // Alignement à droite pour les nbs...
        tmpCase.value = ArrangePrice(stdTVATbl[i].TVA, m_generalCfg);
        tmpLine.cases.push_back(tmpCase);
        tmpCase.value = ArrangePrice(stdTVATbl[i].base, m_generalCfg);
        tmpLine.cases.push_back(tmpCase);
        tmpCase.value = ArrangePrice(stdTVATbl[i].total, m_generalCfg);
        tmpLine.cases.push_back(tmpCase);

        m_TVATbl.lines.push_back(tmpLine); // Ajout au tableau...
        tmpLine.cases.clear(); // On videnge pour la prochaine ligne...
    }

    m_TVATbl.rect.x = 0;
    m_TVATbl.rect.y = 0; // Inconnue pour le moment...
    m_TVATbl.rect.width = ( m_elPos.pageWriteableRect.GetWidth() - m_generalCfg->GetInt("TVArecapTblInterSpace")) / 2;

    m_TVATbl.colsWidth.push_back(int(double(m_TVATbl.rect.width) * 0.2));
    m_TVATbl.colsWidth.push_back(int(double(m_TVATbl.rect.width) * 0.2));
    m_TVATbl.colsWidth.push_back(int(double(m_TVATbl.rect.width) * 0.3));
    m_TVATbl.colsWidth.push_back(int(double(m_TVATbl.rect.width) * 0.3));

    // Hauteur du tableau :
    TableCalculHeightAndPos(m_TVATbl);
}
//------------------------------------------------------------------------------
void DevisPrintout::PrepareFonts()
{
    m_fonts.tableItem = stdColorFontToWxColorFont(m_generalCfg->GetColorFont("tableItem"));
    m_fonts.tableSectionTitle = stdColorFontToWxColorFont(m_generalCfg->GetColorFont("tableSectionTitle"));
    m_fonts.tableTitle = stdColorFontToWxColorFont(m_generalCfg->GetColorFont("tableTitle"));
}
//------------------------------------------------------------------------------
#if defined USE_DEMO_TEXT_IN_PRINTOUT
void DevisPrintout::DrawDemoText()
{
    wxDC *dc = GetDC();

    dc->SetFont(wxFont(48, wxFONTFAMILY_DECORATIVE, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD, false, _T("Arial"), wxFONTENCODING_DEFAULT ));
    dc->SetTextForeground(*wxBLACK);

    mdo::DrawText(dc, wxPoint( m_elPos.pageSize.GetWidth() / 2, m_ItemsTbls[m_ItemsTbls.size() - 1].rect.GetBottom() + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce") / 2), wxString(PROG_FULL_NAME_GETTEXT) + wxString(_T(" - Demo - ")) + wxString(PROG_FULL_NAME_GETTEXT), mdo::vAlign_Center, mdo::hAlign_Center);
}
//------------------------------------------------------------------------------
void DevisPrintout::DrawDemoLogo()
{
    wxDC *dc = GetDC();

    wxPoint pos2(m_elPos.pageSize.GetWidth() * 0.15, m_elPos.pageSize.GetHeight() * 0.6);

    dc->SetFont(wxFont(120, wxFONTFAMILY_DECORATIVE, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD, false, _T("Arial"), wxFONTENCODING_DEFAULT ));
    dc->SetTextForeground(wxColour(237, 237, 239));

    dc->DrawRotatedText(_T("Facture & Devis Maker"), pos2.x, pos2.y, 40);
}
#endif
//------------------------------------------------------------------------------
void DevisPrintout::PrepareRecapTable()
{
    TableClear(m_RecapTbl);

    //m_RecapTbl.doDrawVLine = false; // Pas de ligne verticales
    m_RecapTbl.doDrawVLine = true; // Pas de ligne verticales

    wxDC *dc = GetDC();

    StructTblLine tmpLine;
    StructTblCase tmpCase;

    // Paramétrage des lignes :
    tmpLine.isHline = false;
    tmpLine.interLineSpace = m_generalCfg->GetInt("interLineTVARecapTblSpace");

    // Paramétrage des cases :
    tmpCase.font = &m_fonts.tableItem;
    tmpCase.extraVMarging = 0;
    tmpCase.hAlign = mdo::hAlign_Top;

    // Case 1 :
    tmpCase.vAlign = mdo::vAlign_Left;
    tmpLine.cases.push_back(tmpCase);

    // Case 2 :
    tmpCase.vAlign = mdo::vAlign_Right;
    tmpLine.cases.push_back(tmpCase);

    // Tableau
    if(m_DevisManager->GetDevisCfg()->GetBool("isDumpCharges") || m_DevisManager->GetDevisCfg()->GetBool("isPort"))
    {
        tmpLine.cases[0].value = _("Total sections price (HT) :");
        tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetSectionsPriceHT(), m_generalCfg);
        tmpLine.isHline = false;
        m_RecapTbl.lines.push_back(tmpLine);

        if(m_DevisManager->GetDevisCfg()->GetBool("isPort"))
        {
            tmpLine.cases[0].value = _("Port :");
            tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetPortHT(), m_generalCfg);
            tmpLine.isHline = false;
            m_RecapTbl.lines.push_back(tmpLine);
        }

        if(m_DevisManager->GetDevisCfg()->GetBool("isDumpCharges"))
        {
            tmpLine.cases[0].value = _("Dump charge :");
            tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetDumpChargeHT(), m_generalCfg);
            tmpLine.isHline = false;
            m_RecapTbl.lines.push_back(tmpLine);
        }
    }

    tmpLine.cases[0].value = _("Total Brut (HT) :");
    tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetTotalBrutHT(), m_generalCfg);
    if(m_DevisManager->GetDevisCfg()->GetBool("isDumpCharges") || m_DevisManager->GetDevisCfg()->GetBool("isPort"))
    {
        tmpLine.isHline = true;
    }
    else
    {
        tmpLine.isHline = false;
    }
    m_RecapTbl.lines.push_back(tmpLine);

    tmpLine.cases[0].value = _("Discount rate price :");
    tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetPriceDiscountRateHT(), m_generalCfg);
    tmpLine.isHline = false;
    m_RecapTbl.lines.push_back(tmpLine);

    tmpLine.cases[0].value = _("Discount :");
    tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetDiscountHT(), m_generalCfg);
    tmpLine.isHline = false;
    m_RecapTbl.lines.push_back(tmpLine);

    tmpLine.cases[0].value = _("Total net (HT) :");
    tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetTotalNetHT(), m_generalCfg);
    tmpLine.isHline = false;
    m_RecapTbl.lines.push_back(tmpLine);

    if(m_dvCfg->GetBool("isEcoPart"))
    {
        tmpLine.cases[0].extraVMarging = 70;
        tmpLine.cases[0].value = _("With Eco Part :");
        tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetEcoPartTotalPrice(), m_generalCfg);
        tmpLine.isHline = false;
        m_RecapTbl.lines.push_back(tmpLine);
        tmpLine.cases[0].extraVMarging = 0;
    }

    tmpLine.cases[0].value = _("TVA :");
    tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetTVAPrice(), m_generalCfg);
    tmpLine.isHline = false;
    m_RecapTbl.lines.push_back(tmpLine);

    tmpLine.cases[0].value = _("Total (TTC) :");
    tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetTotalTTC(), m_generalCfg);
    tmpLine.isHline = false;
    m_RecapTbl.lines.push_back(tmpLine);

    tmpLine.cases[0].value = _("Payment on :");
    tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetPaymentOn(), m_generalCfg);
    tmpLine.isHline = true;
    m_RecapTbl.lines.push_back(tmpLine);

    tmpLine.cases[0].value = _("Total to pay :");
    tmpLine.cases[1].value = ArrangePrice(m_DevisManager->GetTotalToPay(), m_generalCfg);
    tmpLine.isHline = false;
    m_RecapTbl.lines.push_back(tmpLine);

    m_RecapTbl.rect.x = (m_elPos.pageWriteableRect.GetWidth() + m_generalCfg->GetInt("TVArecapTblInterSpace")) / 2;
    m_RecapTbl.rect.y = 0; // Inconnue pour le moment...
    m_RecapTbl.rect.width = ( m_elPos.pageWriteableRect.GetWidth() - m_generalCfg->GetInt("TVArecapTblInterSpace")) / 2;

    int chiffreColWidth = TableGetColMaxWidth(m_RecapTbl, 1 );
    m_RecapTbl.colsWidth.push_back(m_RecapTbl.rect.GetWidth() - chiffreColWidth);
    m_RecapTbl.colsWidth.push_back(chiffreColWidth);

    // Hauteur du tableau :
    TableCalculHeightAndPos(m_RecapTbl);
}
//------------------------------------------------------------------------------
void DevisPrintout::TableCalculHeightAndPos(StructTable &table)
{
    // Calcule des vLinesPos...
    table.vLinesPos.clear();

    wxDC *dc = GetDC();

    table.vLinesPos.push_back(table.rect.GetLeft()); // <=="Bordure gauche.
    for(int i = 0; i < table.colsWidth.size(); i++)
    {
        table.vLinesPos.push_back(table.vLinesPos[i] + table.colsWidth[i]);
    }

    // Calcule de la taille...
    int totalHeight = 0;
    for(int i = 0; i < table.lines.size(); i++)
    {
        int height = 0;

        if(table.lines[i].isHline) // Lines
        {
            height += m_generalCfg->GetInt("tableHlineMarging") * 2; // * 2 car une marge de chaque côté...
        }

        height += table.lines[i].interLineSpace;

        // Taille du texte (le plus grand);
        int textHeight = 0;
        for(int casePos = 0; casePos < table.lines[i].cases.size(); casePos++)
        {
            // Utilisation de la font de la case pour les dimensions...
            mdo::DCSelectColorFont(dc, *table.lines[i].cases[casePos].font);

            wxSize tmpSize = mdo::GetTextSize(dc, table.lines[i].cases[casePos].value);

            // Modif du texte s'il ne rentre pas...
            int maxWidth = table.colsWidth[casePos] - m_generalCfg->GetInt("tableVlineMarging") * 2 -  + table.lines[i].cases[casePos].extraVMarging;
            if(tmpSize.GetWidth() > maxWidth)
            {
                table.lines[i].cases[casePos].value = mdo::ArrangeTextForWidth(dc, table.lines[i].cases[casePos].value, maxWidth);

                // On prend la nouvelle taille.
                tmpSize = mdo::GetTextSize(dc, table.lines[i].cases[casePos].value);
            }

            // On prend le plus grand comme taille...
            textHeight = tmpSize.GetHeight() > textHeight ? tmpSize.GetHeight() : textHeight;
        }
        height += textHeight;

        totalHeight += height;
        table.lines[i].height = height;
    }

    table.rect.height = totalHeight;
}
//------------------------------------------------------------------------------
int DevisPrintout::TableGetColMaxWidth(const StructTable &table, const int &col)
{
    wxDC *dc = GetDC();
    int maxWidth = 0;

    for(int i = 0; i < table.lines.size(); i++)
    {
        // Utilisation de la font de la case pour les dimensions...
        mdo::DCSelectColorFont(dc, *table.lines[i].cases[col].font);

        int currentWidth = mdo::GetTextSize(dc, table.lines[i].cases[col].value).GetWidth() + table.lines[i].cases[col].extraVMarging;
        maxWidth = currentWidth > maxWidth ? currentWidth : maxWidth;
    }

    return maxWidth + m_generalCfg->GetInt("tableVlineMarging") * 2; // Ajout de la marge
}
//------------------------------------------------------------------------------
void DevisPrintout::TableCut(const StructTable &table, std::vector<StructTable> &tableList, const int &maxHeight, const int &maxFirstPageHeight, const bool &forceMaxHeightForFirstOnes, const bool &preserveTitle)
{
    tableList.clear();
    bool wantHline = false;

    // Ajout d'un tableau...
    StructTable newTbl;
    newTbl.lines.clear();
    newTbl.colsWidth = table.colsWidth;
    newTbl.doDrawVLine = table.doDrawVLine;
    newTbl.rect = table.rect; // Seul "height" va être modifié.
    newTbl.vLinesPos = table.vLinesPos;
    newTbl.rect.height = 0; // Le nouveau tableau a pour l'instant une hauteur de 0

    int currentMaxHeight;
    if(maxFirstPageHeight >= 0) // Si il faut une hauteur spécial pour la page 1.
    {
        currentMaxHeight = maxFirstPageHeight;
    }
    else
    {
        currentMaxHeight = maxHeight;
    }

    for(int i = 0; i < table.lines.size(); i++)
    {
        // Si encore de la place dans le tableau actuel...
        if(newTbl.rect.height + table.lines[i].height <= currentMaxHeight)
        {
            newTbl.lines.push_back(table.lines[i]); // On ajoute la ligne en cours
            newTbl.rect.height += table.lines[i].height;

            // Line pour titre
            if(wantHline)
            {
                newTbl.lines[newTbl.lines.size() - 1].isHline = true;
                wantHline = false;
            }
        }
        else // Si non ==> Nouveau tableau.
        {
            if(forceMaxHeightForFirstOnes) // Si on met la hauteur max pour les tableau autre que le dernier...
            {
                newTbl.rect.height = currentMaxHeight;
            }
            tableList.push_back(newTbl); // Ajout de l'actuel...

            // Nouveau
            newTbl.lines.clear();
            newTbl.rect.height = 0; // Le nouveau tableau a pour l'instant une hauteur de 0

            // Rajout du titre
            if(preserveTitle && table.lines.size() >= 2)
            {
                newTbl.lines.push_back(table.lines[0]);
                newTbl.rect.height += table.lines[0].height;

                wantHline = table.lines[1].isHline;
            }

            // Hauteur (plus le premier tableau)
            currentMaxHeight = maxHeight;
        }
    }

    tableList.push_back(newTbl); // Ajout de l'actuel...
}
//------------------------------------------------------------------------------
void DevisPrintout::TableClear(StructTable &table)
{
    table.colsWidth.clear();
    table.lines.clear();
    table.vLinesPos.clear();
}
//------------------------------------------------------------------------------
void DevisPrintout::PreparePreInfo()
{
    wxDC *dc = GetDC();
    int width, height;

    mdo::Sizer *sizer;

    // PRE INFO :
    sizer = m_preInfoSizer;

    sizer->LoadFromCfg();
    if(sizer->GetWidthProportion() == -1)
    {
        width = m_elPos.pageWriteableRect.GetWidth();
    }
    else
    {
        width = m_elPos.pageWriteableRect.GetWidth() * sizer->GetWidthProportion();
    }

    if(sizer->GetHeightProportion() == -1)
    {
        height = m_elPos.pageWriteableRect.GetHeight();
    }
    else
    {
        height = m_elPos.pageWriteableRect.GetHeight() * sizer->GetHeightProportion();
    }

    sizer->CalculSizeReal(dc, width, height);
}
//------------------------------------------------------------------------------
void DevisPrintout::PreparePostInfo()
{
    wxDC *dc = GetDC();
    int width, height;

    mdo::Sizer *sizer;

    // POST INFO :
    sizer = m_postInfoSizer;

    sizer->LoadFromCfg();
    if(sizer->GetWidthProportion() == -1)
    {
        width = m_elPos.pageWriteableRect.GetWidth();
    }
    else
    {
        width = m_elPos.pageWriteableRect.GetWidth() * sizer->GetWidthProportion();
    }

    if(sizer->GetHeightProportion() == -1)
    {
        height = m_elPos.pageWriteableRect.GetHeight();
    }
    else
    {
        height = m_elPos.pageWriteableRect.GetHeight() * sizer->GetHeightProportion();
    }

    sizer->CalculSizeReal(dc, width, height);
}
//------------------------------------------------------------------------------
DevisPrintout::~DevisPrintout()
{
    delete m_preInfoSizer;
    delete m_postInfoSizer;
}
//------------------------------------------------------------------------------
void DevisPrintout::PrepareItemsTable()
{
    wxDC *dc = GetDC();
    StructTable itemsTbl;
    TableClear(itemsTbl);

    // TVAid :
    std::vector<StructTVALine> TVATbl = m_DevisManager->GetTVATbl();
    bool showTVAid = false;

    if(m_generalCfg->GetBool("showTVAidInItemsTbl"))
    {
        if(m_generalCfg->GetBool("showTVAidInItemsTbl_OnlyIfMoreThanTVAid"))
        {
            if(TVATbl.size() > 1)
            {
                showTVAid = true;
            }
        }
        else
        {
            showTVAid = true;
        }
    }


    // Param Table
    itemsTbl.doDrawVLine = true;

    StructTblLine tmpLine;
    StructTblCase tmpCase;

    // Paramétrage des lignes :
    tmpLine.isHline = false; // Pas de ligne au dessus du titre

    // Préparation des cases
    tmpCase.extraVMarging = 0;

    // Case titre :
    tmpCase.hAlign = mdo::hAlign_Center;
    tmpCase.vAlign = mdo::vAlign_Center;
    tmpLine.interLineSpace = m_generalCfg->GetInt("titleLineHeight");
    tmpCase.font = &m_fonts.tableTitle;

	tmpCase.value = _("Reference");
	tmpLine.cases.push_back(tmpCase);
	tmpCase.value = _("Item name");
	tmpLine.cases.push_back(tmpCase);
	tmpCase.value = _("Item price");
	tmpLine.cases.push_back(tmpCase);
    if(m_DevisManager->GetDevisCfg()->GetBool("isEcoPart"))
    {
        tmpCase.value = _("Eco Part");
        tmpLine.cases.push_back(tmpCase);
    }
	tmpCase.value = _("Quantity");
	tmpLine.cases.push_back(tmpCase);
	tmpCase.value = _("Price");
	tmpLine.cases.push_back(tmpCase);
	if(showTVAid)
	{
        tmpCase.value = _("T");
        tmpLine.cases.push_back(tmpCase);
	}

	itemsTbl.lines.push_back(tmpLine); // Ajout au tableau...
	tmpLine.cases.clear(); // On videnge pour la prochaine ligne...


    wxString str;
    std::vector<std::string> SectionsIDs = m_DevisManager->GetSectionsList(); // Chargement des sections

    for( size_t posSection = 0; posSection < SectionsIDs.size(); posSection++) // Boucle des piéces
    {
        if(m_generalCfg->GetBool("isPrintSections"))
        {
            // Case section title
            tmpCase.hAlign = mdo::hAlign_Bottom; // En bas pour laisser un espace avec le texte avant...
            tmpCase.vAlign = mdo::vAlign_Left;
            tmpCase.font = &m_fonts.tableSectionTitle;
            if(posSection == 0) // Juste après le titre ==> separation.
            {
                tmpLine.isHline = true;
                tmpLine.interLineSpace = 0; // Premiére ligne : pas d'espace avec le haut du tableau.
            }
            else
            {
                tmpLine.isHline = false;
                tmpLine.interLineSpace = m_generalCfg->GetInt("interSectionSpace");
            }

            for(int i = 0; i < itemsTbl.lines[0].cases.size(); i++) // Nb de colonnes
            {
                if(i == ITEM_NAME_COL_POS) // Name
                {
                    tmpCase.extraVMarging = 70;
                    tmpCase.value = StrStdToWx(SectionsIDs[posSection]);
                    tmpLine.cases.push_back(tmpCase);
                    tmpCase.extraVMarging = 0;
                }
                else // Autre ==> Vide
                {
                    tmpCase.value = _T("");
                    tmpLine.cases.push_back(tmpCase);
                }
            }

            itemsTbl.lines.push_back(tmpLine); // Ajout au tableau...
            tmpLine.cases.clear(); // On videnge pour la prochaine ligne...
        }

        // Case item
        tmpCase.hAlign = mdo::hAlign_Top; // En bas pour laisser un espace avec le texte avant...
        tmpLine.isHline = false;
        tmpCase.font = &m_fonts.tableItem;
        tmpLine.interLineSpace = m_generalCfg->GetInt("interItemsSpace");

        std::vector<std::string> ItemsIDs = m_DevisManager->GetItemsList(posSection); // Chargement des items

        tmpLine.cases.clear(); // On videnge pour la prochaine ligne...

        for( size_t posItem = 0; posItem < ItemsIDs.size(); posItem++) // Boucle des items
        {
            if(posSection == 0 && posItem == 0 && !m_generalCfg->GetBool("isPrintSections")) // Juste après le titre ==> separation (si pas de section, car si non, HLine déjà mise).
            {
                tmpLine.isHline = true;
            }
            else
            {
                tmpLine.isHline = false;
            }

            // Ref
            tmpCase.vAlign = mdo::vAlign_Center;
            tmpCase.value = StrStdToWx(ItemsIDs[posItem]);
            tmpLine.cases.push_back(tmpCase);

            // Name
            tmpCase.vAlign = mdo::vAlign_Left;
            tmpCase.value = StrStdToWx(m_DevisManager->GetItemName(ItemsIDs[posItem]));
            tmpLine.cases.push_back(tmpCase);

            // ItemPrice
            tmpCase.vAlign = mdo::vAlign_Right;
            tmpCase.value = ArrangePrice(m_DevisManager->GetItemPriceHT(ItemsIDs[posItem]), m_generalCfg);
            tmpLine.cases.push_back(tmpCase);

            // Eco Part
            if(m_DevisManager->GetDevisCfg()->GetBool("isEcoPart"))
            {
                tmpCase.vAlign = mdo::vAlign_Right;
                tmpCase.value = ArrangePrice(m_DevisManager->GetItemEcoPart(ItemsIDs[posItem]), m_generalCfg);
                tmpLine.cases.push_back(tmpCase);
            }

            // Quantity
            tmpCase.vAlign = mdo::vAlign_Right;
            tmpCase.value = wxNbToStr(m_DevisManager->GetItemQuantity(posSection, posItem));
            tmpLine.cases.push_back(tmpCase);

            // Price
            tmpCase.vAlign = mdo::vAlign_Right;
            tmpCase.value = ArrangePrice(m_DevisManager->GetItemPriceWithQuantityHT(posSection, posItem), m_generalCfg);
            tmpLine.cases.push_back(tmpCase);

            // showTVAid
            if(showTVAid)
            {
                tmpCase.vAlign = mdo::vAlign_Center;
                tmpCase.value = StrStdToWx(m_DevisManager->GetItemTVAid(posSection, posItem, TVATbl));
                tmpLine.cases.push_back(tmpCase);
            }
            // Fin
            itemsTbl.lines.push_back(tmpLine); // Ajout au tableau...
            tmpLine.cases.clear(); // On videnge pour la prochaine ligne...
        }
    }

    itemsTbl.rect.x = 0;
    itemsTbl.rect.y = 0;
    itemsTbl.rect.width = m_elPos.pageWriteableRect.GetWidth();

    // Cols Width
    int totalSize = 0;
    for(int i = 0; i < itemsTbl.lines[0].cases.size(); i++) // Nb de colonnes
    {
        if(i != ITEM_NAME_COL_POS) // Si on n'est pas à la colonne name...
        {
            itemsTbl.colsWidth.push_back(TableGetColMaxWidth(itemsTbl, i));
        }
        else
        {
            itemsTbl.colsWidth.push_back(0); // Name ==> à calculer (le reste)
        }

        totalSize += itemsTbl.colsWidth[i];
    }

    // On donne la taille à la collone name (le reste)
    if(m_elPos.pageWriteableRect.GetWidth() - totalSize > 0)
    {
        itemsTbl.colsWidth[ITEM_NAME_COL_POS] = itemsTbl.rect.GetWidth() - totalSize; // Name ==> place restante.
    }
    else // Si pas possible (plus de place) ==> Valeurs par defaut.
    {
        itemsTbl.colsWidth.clear();

        // On divise.
        for(int i = 0; i < itemsTbl.lines[0].cases.size(); i++) // Nb de colonnes
        {
            itemsTbl.colsWidth.push_back(m_elPos.pageWriteableRect.GetWidth() / itemsTbl.lines[0].cases.size());
        }
    }

    // Hauteur du tableau :
    TableCalculHeightAndPos(itemsTbl);

    // Découpage...
    TableCut(itemsTbl, m_ItemsTbls, m_elPos.pageWriteableRect.GetHeight(), m_firstPageTableHeight, m_generalCfg->GetBool("isTableMaxHeight"));

    // Si le tableau est avec le preInfo
    m_ItemsTbls[0].rect.SetY(m_elPos.firstItemTblPos.y);

    // Nb de page...
    m_nbPage += m_ItemsTbls.size();
}
//------------------------------------------------------------------------------
void DevisPrintout::PrepareElementsPos()
{
    // Page Real Size
    int pageRealWidthPx = 0;
    int pageRealHeightPx = 0;
    if( m_generalCfg->GetBool("usePrinterMarging"))
    {
        GetPageSizePixels(&pageRealWidthPx, &pageRealHeightPx);
    }
    else
    {
        wxRect paperRect = GetPaperRectPixels();
        pageRealWidthPx = paperRect.GetWidth();
        pageRealHeightPx = paperRect.GetHeight();
    }

    // Page Size
    // Si page plus haute que large (portrait) ==> On augmente la hauteur par rapport au A4
    if(pageRealHeightPx >= pageRealWidthPx)
    {
        m_elPos.pageSize.SetWidth(PAGE_WIDTH);
        m_elPos.pageSize.SetHeight(int(double(m_elPos.pageSize.GetWidth()) / (double(pageRealWidthPx) / double(pageRealHeightPx))));
    }
    else // Si non, on augmente la largeur.
    {
        m_elPos.pageSize.SetHeight(PAGE_HEIGHT);
        m_elPos.pageSize.SetWidth(int(double(m_elPos.pageSize.GetHeight()) * (double(pageRealWidthPx) / double(pageRealHeightPx))));
    }

    // pageWriteableRect
    wxPoint pageWriteableLeftTop;
    pageWriteableLeftTop.x = m_generalCfg->GetInt("pageMargingLeft");
    pageWriteableLeftTop.y = m_generalCfg->GetInt("pageMargingTop");

    wxPoint pageWriteableRightBottom;
    pageWriteableRightBottom.x = m_elPos.pageSize.GetWidth() - m_generalCfg->GetInt("pageMargingRight");
    pageWriteableRightBottom.y = m_elPos.pageSize.GetHeight() - m_generalCfg->GetInt("pageMargingBottom");

    m_elPos.pageWriteableRect = wxRect( pageWriteableLeftTop, pageWriteableRightBottom );
}
//------------------------------------------------------------------------------
void DevisPrintout::PreparePage()
{
    // Param de la page
    if( m_generalCfg->GetBool("usePrinterMarging"))
    {
        FitThisSizeToPage(m_elPos.pageSize);
    }
    else
    {
        FitThisSizeToPaper(m_elPos.pageSize);
    }

    // On change l'origine du papier en fonction des marges.
    OffsetLogicalOrigin(m_elPos.pageWriteableRect.x, m_elPos.pageWriteableRect.y);
}
//------------------------------------------------------------------------------
void DevisPrintout::PrepareAll()
{
    m_nbPage = 0;

    PrepareElementsPos();
    PreparePage();
    PrepareFonts();
    PreparePreInfo();

    // Pre info :
    // Page 1 : tableau ?
    int usedYPlacePreInfo = m_preInfoSizer->GetSizeReal().GetHeight();
    int leftYPlacePreInfo = m_elPos.pageWriteableRect.GetHeight() - usedYPlacePreInfo;

    m_elPos.preInfoPos.x = 0;
    m_elPos.preInfoPos.y = 0;
    m_elPos.preInfoPos.page = 0;

    m_elPos.firstItemTblPos.x = 0;
    if(leftYPlacePreInfo > m_generalCfg->GetInt("titleLineHeight")) // Il y a la place
    {
        m_firstPageTableHeight = leftYPlacePreInfo;

        m_elPos.firstItemTblPos.y = usedYPlacePreInfo;
        m_elPos.firstItemTblPos.page = 0;
    }
    else
    {
        m_firstPageTableHeight = m_elPos.pageWriteableRect.GetHeight();

        m_elPos.firstItemTblPos.y = 0;
        m_elPos.firstItemTblPos.page = 1;
        m_nbPage ++;
    }

    PrepareItemsTable();
    PrepareRecapTable();
    PrepareTVATable();

    // Page (last) : recap + tva rentre ?
    int usedYPlaceTvaRecap = m_ItemsTbls[m_ItemsTbls.size() - 1].rect.GetBottom();
    int leftYPlaceTvaRecap = m_elPos.pageWriteableRect.GetHeight() - usedYPlaceTvaRecap;
    int neededHeightForRecapTvaTbl = std::max(m_TVATbl.rect.GetHeight(), m_RecapTbl.rect.GetHeight()) + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");

    bool isNewPageForTvaRecapTbl = false;

    if(leftYPlaceTvaRecap >= neededHeightForRecapTvaTbl) // Si il y a de la place sur la page...
    {
        isNewPageForTvaRecapTbl = false;

        m_elPos.tvaRecapTblPos.x = 0;
        m_elPos.tvaRecapTblPos.y = usedYPlaceTvaRecap;
        m_elPos.tvaRecapTblPos.page = m_nbPage - 1;

        int maxHeight = std::max(m_TVATbl.rect.GetHeight(), m_RecapTbl.rect.GetHeight());

        //m_tableOnLastPage = true;
        if(m_generalCfg->GetBool("isTableMaxHeight")) // Si on demande une taille max pour les tbl ==> Il faut agrandir ItemTbl sur la derniére page (avant ==> fait pas TableCut())
        {
            m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height = m_elPos.pageWriteableRect.GetHeight() - m_ItemsTbls[m_ItemsTbls.size() - 1].rect.GetY() - neededHeightForRecapTvaTbl - m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");

            m_RecapTbl.rect.SetHeight(maxHeight);
            m_TVATbl.rect.SetHeight(maxHeight);
        }

        int recapTvaPos = m_ItemsTbls[m_ItemsTbls.size() - 1].rect.GetBottom() + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");
        m_RecapTbl.rect.SetY(recapTvaPos);
        m_TVATbl.rect.SetY(recapTvaPos);
    }
    else // Si non, on ajoute une page
    {
        isNewPageForTvaRecapTbl = true;

        m_nbPage++;
        m_RecapTbl.rect.SetY(0);
        m_TVATbl.rect.SetY(0);

        m_elPos.tvaRecapTblPos.x = 0;
        m_elPos.tvaRecapTblPos.y = 0;
        m_elPos.tvaRecapTblPos.page = m_nbPage - 1;

        if(m_generalCfg->GetBool("isTableMaxHeight")) // Si on demande une taille max pour les tbl ==> Il faut agrandir ItemTbl sur la derniére page (avant ==> faut pas TableCut())
        {
            m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height = m_firstPageTableHeight;
        }

        usedYPlaceTvaRecap = 0; // Plus de place utilisée avant le tva & recap tbl (pour post info)
    }

    PreparePostInfo();


    // Post info ==> Nouvelle page ?
    int usedYPlacePostInfo = usedYPlaceTvaRecap + neededHeightForRecapTvaTbl;
    int leftYPlacePostInfo = m_elPos.pageWriteableRect.GetHeight() - usedYPlacePostInfo;
    int neededHeightForPostInfo = m_postInfoSizer->GetSizeReal().GetHeight();

    if(leftYPlacePostInfo >= neededHeightForPostInfo) // Si il y a de la place sur la page...
    {
        int recapTvaPos;
        int RecapTvaHeight = std::max(m_TVATbl.rect.GetHeight(), m_RecapTbl.rect.GetHeight());

        if( !isNewPageForTvaRecapTbl ) // Si on a pas ajouté de page pour le TVA tbl
        {
            if(m_generalCfg->GetBool("isTableMaxHeight")) // Si on demande une taille max pour les tbl ==> Il faut agrandir ItemTbl sur la derniére page (avant ==> faut pas TableCut())
            {
                m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height = m_elPos.pageWriteableRect.GetHeight() - m_ItemsTbls[m_ItemsTbls.size() - 1].rect.GetY() - neededHeightForRecapTvaTbl - neededHeightForPostInfo - m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");
            }

            // On colle les recap & tva tbl un max
            recapTvaPos = m_ItemsTbls[m_ItemsTbls.size() - 1].rect.GetBottom() + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");
            m_RecapTbl.rect.SetY(recapTvaPos);
            m_TVATbl.rect.SetY(recapTvaPos);
        }
        else
        {
            recapTvaPos = m_RecapTbl.rect.GetY();
        }

        m_elPos.postInfoPos.x = 0;
        m_elPos.postInfoPos.y = recapTvaPos + RecapTvaHeight;
        m_elPos.postInfoPos.page = m_nbPage - 1;
    }
    else // Si non, on ajoute une page
    {
        m_nbPage++;
        m_elPos.postInfoPos.x = 0;
        m_elPos.postInfoPos.y = 0;
        m_elPos.postInfoPos.page = m_nbPage - 1;
    }





    // Page (last) : recap + tva rentre ?
    /*usedYPlace = m_postInfoSizer->GetSizeReal()->GetHeight() + m_ItemsTbls[m_ItemsTbls.size() - 1].rect.GetHeight() + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");
    leftYPlace = m_elPos.pageWriteableRect.GetHeight() - usedYPlace;
    int neededHeightForRecapTvaTbl = max(m_TVATbl.rect.GetHeight(), m_RecapTbl.rect.GetHeight());*/

    /*if(leftYPlace > 0)
    {
        firstPageTableHeight = m_elPos.pageWriteableRect.GetHeight() - usedYPlace;
    }

    m_nbPage = m_ItemsTbls.size() + 1;

    // Pos du tbl recap et TVA ==> Place requise
    int usedYPlace = m_ItemsTbls[m_ItemsTbls.size() - 1].rect.GetHeight() + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");
    int leftYPlace = m_elPos.pageWriteableRect.GetHeight() - usedYPlace;
    // On prend la taille la plus grande.
    int neededHeight = (m_TVATbl.rect.GetHeight() > m_RecapTbl.rect.GetHeight()) ? m_TVATbl.rect.GetHeight() : m_RecapTbl.rect.GetHeight();

    if(m_generalCfg->GetBool("isTableMaxHeight")) // Si on demande une taille max pour les tbl ==> TVATbl & RecapTbl ==> même taille
    {
        m_TVATbl.rect.height = neededHeight;
        m_RecapTbl.rect.height = neededHeight;
    }*/


    /*if(leftYPlace >= neededHeightForRecapTvaTbl) // Si il y a de la place sur la page...
    {
        m_tableOnLastPage = true;
        if(m_generalCfg->GetBool("isTableMaxHeight")) // Si on demande une taille max pour les tbl ==> Il faut agrandir ItemTbl sur la derniére page (avant ==> faut pas TableCut())
        {
            m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height = m_elPos.pageWriteableRect.GetHeight() - neededHeight - m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");
        }

        m_RecapTbl.rect.SetY(m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce"));
        m_TVATbl.rect.SetY(m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce"));

        usedYPlace += neededHeightForRecapTvaTbl;
    }
    else // Si non, on ajoute une page
    {
        m_tableOnLastPage = false;
        m_nbPage++;
        m_RecapTbl.rect.SetY(0);
        m_TVATbl.rect.SetY(0);
        m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height = m_elPos.pageWriteableRect.GetHeight();

        usedYPlace = neededHeightForRecapTvaTbl;
        leftYPlace = m_elPos.pageWriteableRect.GetHeight() - usedYPlace;
    }

    // Post info ==> Nouvelle page ?
    int neededHeightForPostInfo = m_postInfoSizer->GetSizeReal().GetHeight();

    if(leftYPlace >= neededHeightForPostInfo) // Si il y a de la place sur la page...
    {
        m_tableOnLastPage = true;
        if(m_generalCfg->GetBool("isTableMaxHeight")) // Si on demande une taille max pour les tbl ==> Il faut agrandir ItemTbl sur la derniére page (avant ==> faut pas TableCut())
        {
            m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height = m_elPos.pageWriteableRect.GetHeight() - neededHeight - m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce");
        }

        m_RecapTbl.rect.SetY(m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce"));
        m_TVATbl.rect.SetY(m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height + m_generalCfg->GetInt("TVArecapTbANDitemsInterSapce"));

        usedYPlace += neededHeightForRecapTvaTbl;
    }
    else // Si non, on ajoute une page
    {
        m_tableOnLastPage = false;
        m_nbPage++;
        m_RecapTbl.rect.SetY(0);
        m_TVATbl.rect.SetY(0);
        m_ItemsTbls[m_ItemsTbls.size() - 1].rect.height = m_elPos.pageWriteableRect.GetHeight();

        usedYPlace = neededHeightForRecapTvaTbl;
        leftYPlace = m_elPos.pageWriteableRect.GetHeight() - usedYPlace;
    }*/
}
//------------------------------------------------------------------------------
DevisPrintout::DevisPrintout(wxString title, DevisManager *devisManager) : wxPrintout(title)
{
    m_DevisManager = devisManager;
    m_dvCfg = m_DevisManager->GetDevisCfg();
    m_generalCfg = &m_DevisManager->GetGeneralCfg()->GetGeneralCfg();

    // Mdo
    m_postInfoSizer = new mdo::Sizer(m_generalCfg, m_dvCfg, NULL, "postInfo");
    m_preInfoSizer = new mdo::Sizer(m_generalCfg, m_dvCfg, NULL, "preInfo");
}
//------------------------------------------------------------------------------
void DevisPrintout::DrawPreInfo()
{
    wxDC *dc = GetDC();
    m_preInfoSizer->Draw(dc, wxPoint(m_elPos.preInfoPos.x, m_elPos.preInfoPos.y) );
}
//------------------------------------------------------------------------------
void DevisPrintout::DrawPostInfo()
{
    wxDC *dc = GetDC();
    m_postInfoSizer->Draw(dc, wxPoint(m_elPos.postInfoPos.x, m_elPos.postInfoPos.y));
}
//------------------------------------------------------------------------------
void DevisPrintout::DrawTable(const StructTable &table)
{
    int nextYpos = table.rect.GetTop();

    wxDC *dc = GetDC();

    // Ctrl
    for(int linePos = 0; linePos < table.lines.size(); linePos++)
    {
    	int textYPosDecal = 0;

        if(table.lines[linePos].isHline) // Lines
        {
            textYPosDecal = m_generalCfg->GetInt("tableHlineMarging");
            dc->DrawLine(table.rect.GetLeft(), nextYpos + textYPosDecal, table.rect.GetRight(), nextYpos + textYPosDecal);
            textYPosDecal = m_generalCfg->GetInt("tableHlineMarging");
        }

        for(int casePos = 0; casePos < table.lines[linePos].cases.size(); casePos++)
        {
        	wxPoint drawPos(0, 0);
			switch(table.lines[linePos].cases[casePos].vAlign)
			{
				case mdo::vAlign_Left:
					drawPos.x = table.vLinesPos[casePos] + m_generalCfg->GetInt("tableVlineMarging") + table.lines[linePos].cases[casePos].extraVMarging;
					break;
				case mdo::vAlign_Center:
					drawPos.x = table.vLinesPos[casePos] + table.colsWidth[casePos] / 2;
					break;
				case mdo::vAlign_Right:
					drawPos.x = table.vLinesPos[casePos + 1] - m_generalCfg->GetInt("tableVlineMarging") - table.lines[linePos].cases[casePos].extraVMarging;
					break;
                default:
                    throw ticpp::Exception("Unknown valign !");
                    break;
			}

			switch(table.lines[linePos].cases[casePos].hAlign)
			{
				case mdo::hAlign_Top:
	 				drawPos.y = nextYpos + textYPosDecal;
					break;
				case mdo::hAlign_Center:
					drawPos.y = nextYpos + textYPosDecal + table.lines[linePos].height / 2;
					break;
				case mdo::hAlign_Bottom:
					drawPos.y = nextYpos + textYPosDecal + table.lines[linePos].height;
					break;
                default:
                    throw ticpp::Exception("Unknown halign !");
                    break;
			}
			mdo::DCSelectColorFont(dc, *table.lines[linePos].cases[casePos].font);
		    mdo::DrawText(dc, drawPos, table.lines[linePos].cases[casePos].value, table.lines[linePos].cases[casePos].vAlign, table.lines[linePos].cases[casePos].hAlign);
        }

        nextYpos+= table.lines[linePos].height;
    }

    // Vlines :
    if(table.doDrawVLine)
    {
        for(int i = 1; i < table.vLinesPos.size() - 1; i++) // On ne trace pas la premiére et la derniére ligne ==> On le fait après.
        {
            dc->DrawLine(table.vLinesPos[i], table.rect.GetTop(), table.vLinesPos[i], table.rect.GetBottom());
        }
    }

    // Bordures
    dc->DrawRoundedRectangle(table.rect.x, table.rect.y, table.rect.width, table.rect.height, m_generalCfg->GetInt("arcSize"));
}
//------------------------------------------------------------------------------
bool DevisPrintout::OnPrintPage(int page)
{
    try
    {
        wxDC *dc = GetDC();

        if(dc)
        {
            // Les pages commence à 0...
            page--;

            // Param de la page
            PreparePage();

            // Param du DC
            dc->SetBrush(*wxTRANSPARENT_BRUSH);
            dc->SetPen(wxPen( stdColorToWxColor(m_generalCfg->GetColor("tableColor")) ));

            #if defined USE_DEMO_TEXT_IN_PRINTOUT // En premier pour le fond.
            if(page >= m_elPos.firstItemTblPos.page && page <= m_elPos.firstItemTblPos.page + m_ItemsTbls.size() - 1)
            {
                DrawDemoLogo();
            }
            #endif

            if(page == m_elPos.preInfoPos.page)
            {
                DrawPreInfo();
            }

            if(page >= m_elPos.firstItemTblPos.page && page <= m_elPos.firstItemTblPos.page + m_ItemsTbls.size() - 1)
            {
                #if defined USE_DEMO_TEXT_IN_PRINTOUT
                DrawDemoLogo();

                if(page == m_elPos.firstItemTblPos.page + m_ItemsTbls.size() - 1) // Derniére page de tableau ==> Texte de demo
                {
                    DrawDemoText();
                }
                #endif

                DrawTable(m_ItemsTbls[page - m_elPos.firstItemTblPos.page]);
            }

            if(page == m_elPos.tvaRecapTblPos.page)
            {
                DrawTable(m_TVATbl);
                DrawTable(m_RecapTbl);
            }

            if(page == m_elPos.postInfoPos.page)
            {
                DrawPostInfo();
            }


            /*if(page == 1)
            {
                DrawPreInfo();
            }
            else if(page > 1 && page < m_nbPage)
            {
                int tablePage = page - 2; // Les tbls commence à 0 et il y a le header...
                DrawTable(m_ItemsTbls[tablePage]);
            }
            else if(page == m_nbPage) // Derniére page ==> Recap + ?Tbl
            {
                if(m_tableOnLastPage)
                {
                    int tablePage = page - 2; // Les tbls commence à 0 et il y a le header...
                    DrawTable(m_ItemsTbls[tablePage]);
                }

                DrawTable(m_RecapTbl);
                DrawTable(m_TVATbl);
            }
            else
            {
                return false;
            }*/

            // Liberartion du DC
            dc->SetBrush(wxNullBrush);
            dc->SetPen(wxNullPen);
        }
        else
        {
            return false;
        }
    }
    catch(const std::exception& e) //Rattrape toutes les exceptions
    {
        wxMessageBox(_("Error while printing : ") + wxString(_T("\n")) + StrStdToWx(e.what()), _("Error..."),  wxOK | wxICON_ERROR);
        return false;
    }

    return true;
}
//------------------------------------------------------------------------------

bool DevisPrintout::HasPage(int page)
{
    if(page >= 1 && page <= m_nbPage)
    {
        return true;
    }
    else
    {
        return false;
    }
}
//------------------------------------------------------------------------------

bool DevisPrintout::OnBeginDocument(int startPage, int endPage)
{
    if(wxPrintout::OnBeginDocument(startPage, endPage))
    {
        return true;
    }
    else
    {
        return false;
    }
}
//------------------------------------------------------------------------------

void DevisPrintout::GetPageInfo(int *minPage, int *maxPage, int *selPageFrom, int *selPageTo)
{
    if(m_DevisManager->GetSectionsList().size() != 0)
    {
        try
        {
            PrepareAll();
        }
        catch(const std::exception& e) //Rattrape toutes les exceptions
        {
            wxMessageBox(_("Error while printing : ") + wxString(_T("\n")) + StrStdToWx(e.what()), _("Error..."),  wxOK | wxICON_ERROR);
            *minPage = 0;
            *maxPage = 0;
            *selPageFrom = 0;
            *selPageTo = 0;
        }

        *minPage = 1;
        *maxPage = m_nbPage;
        *selPageFrom = 1;
        *selPageTo = m_nbPage;
    }
    else
    {
        *minPage = 1;
        *maxPage = 1;
        *selPageFrom = 1;
        *selPageTo = 1;
    }
}
//------------------------------------------------------------------------------
